from dbt_af.conf.config import (
    Config,
    CustomAfCallbacksConfig,
    DbtDefaultTargetsConfig,
    DbtProjectConfig,
    K8sConfig,
    MCDIntegrationConfig,
    RetriesConfig,
    RetryPolicy,
    TableauIntegrationConfig,
)

__all__ = [
    'Config',
    'DbtDefaultTargetsConfig',
    'DbtProjectConfig',
    'K8sConfig',
    'MCDIntegrationConfig',
    'TableauIntegrationConfig',
    'CustomAfCallbacksConfig',
    'RetriesConfig',
    'RetryPolicy',
]



import json
from pathlib import Path

import pytest
import yaml

from dbt_af.builder.dbt_af_builder import DbtAfGraph
from dbt_af.conf import Config, DbtDefaultTargetsConfig, DbtProjectConfig


def pytest_addoption(parser):
    parser.addoption('--manifest_path', action='store', default='.', help='Path to manifest.json')
    parser.addoption('--profiles_path', action='store', default='.', help='Path to profiles.yml')
    parser.addoption('--dbt_project_path', action='store', default='.', help='Path to dbt_project.yml')
    parser.addoption('--models_path', action='store', default='.', help='Path to models')
    parser.addoption('--etl_name', action='store', default=None, help='List of ETLs to test')
    parser.addoption('--target', action='store', default='dev', help='Default target to use')


@pytest.fixture(scope='session')
def manifest_path(pytestconfig) -> Path:
    manifest_path = pytestconfig.getoption('manifest_path')
    return Path(manifest_path) if manifest_path.endswith('manifest.json') else Path(manifest_path) / 'manifest.json'


@pytest.fixture(scope='session')
def profiles_path(pytestconfig) -> Path:
    profiles_path = pytestconfig.getoption('profiles_path')
    return Path(profiles_path) if profiles_path.endswith('profiles.yml') else Path(profiles_path) / 'profiles.yml'


@pytest.fixture(scope='session')
def dbt_project_path(pytestconfig) -> Path:
    dbt_project_path = pytestconfig.getoption('dbt_project_path')
    return (
        Path(dbt_project_path)
        if dbt_project_path.endswith('dbt_project.yml')
        else Path(dbt_project_path) / 'dbt_project.yml'
    )


@pytest.fixture(scope='session')
def models_path(pytestconfig) -> Path:
    return Path(pytestconfig.getoption('models_path'))


@pytest.fixture(scope='session')
def etl_name(pytestconfig):
    return pytestconfig.getoption('etl_name')


@pytest.fixture(scope='session')
def target(pytestconfig):
    return pytestconfig.getoption('target')


@pytest.fixture
def config(models_path, dbt_project_path, profiles_path, target):
    return Config(
        dbt_project=DbtProjectConfig(
            dbt_project_name='dwh',
            dbt_models_path=models_path,
            dbt_project_path=dbt_project_path,
            dbt_profiles_path=profiles_path,
            dbt_target_path=dbt_project_path.parent,
            dbt_log_path=dbt_project_path.parent,
            dbt_schema='schema',
        ),
        dbt_default_targets=DbtDefaultTargetsConfig(default_target=target),
    )


@pytest.fixture
def manifest(manifest_path) -> dict:
    with open(manifest_path, 'r') as f:
        return json.load(f)


@pytest.fixture
def profiles(profiles_path) -> dict:
    with open(profiles_path, 'r') as f:
        return yaml.safe_load(f)


@pytest.fixture
def dbt_project(dbt_project_path):
    with open(dbt_project_path, 'r') as f:
        return yaml.safe_load(f)


@pytest.fixture
def project_profile_name(dbt_project):
    return dbt_project['profile']


@pytest.fixture
def dbt_af_graph(manifest, profiles, project_profile_name, config):
    return DbtAfGraph.from_manifest(manifest, profiles, project_profile_name, etl_service_name='', config=config)



from .callbacks import prepare_mcd_callbacks  # noqa
from .dbt_core import send_dbt_artifacts_to_montecarlo  # noqa

__all__ = [
    'prepare_mcd_callbacks',
    'send_dbt_artifacts_to_montecarlo',
]



import collections
import typing as tp

from dbt_af.parser.dbt_node_model import DbtNode
from dbt_af.parser.dbt_source_model import DbtSource


class DbtModelPathGraph:
    def __init__(self, nodes: tp.List[DbtNode], sources: tp.List[DbtSource]):
        self.dbt_nodes = nodes
        self.dbt_sources = sources

        self.node_to_path = {}
        self.path_to_path = collections.defaultdict(set)
        self.full_path_to_path = collections.defaultdict(set)

        self._build_path_dependencies()

    @classmethod
    def from_manifest(cls, manifest: dict) -> 'DbtModelPathGraph':
        nodes = []
        for node_info in manifest['nodes'].values():
            node = DbtNode(**node_info)
            if node.resource_type in ('model', 'snapshot', 'test'):
                nodes.append(node)

        sources = [DbtSource(**source_info) for source_info in manifest['sources'].values()]

        return cls(nodes, sources)

    def _build_path_dependencies(self):
        # step1 Create dict node->path

        self.node_to_path = {node.unique_id: node.original_file_path_dirname for node in self.dbt_nodes}
        self.node_to_path.update({source.unique_id: source.original_file_path_dirname for source in self.dbt_sources})

        # step2 Create dict path (depends on) set of path
        for node in self.dbt_nodes:
            for dep in node.depends_on:
                dep_path = self.node_to_path[dep]
                if dep_path != node.original_file_path_dirname:
                    self.path_to_path[node.original_file_path_dirname].add(self.node_to_path[dep])
            for dep in node.depends_on_sources:
                dep_path = self.node_to_path[dep]
                if dep_path != node.original_file_path_dirname:
                    self.path_to_path[node.original_file_path_dirname].add(self.node_to_path[dep])

        # step3 Create dict path (depends on) set of all pathes
        for path, dep_path in self.path_to_path.copy().items():
            # all pathes form step2 above
            self.full_path_to_path[path].update(dep_path)

            # create queue with pathes we depends on
            queue = collections.deque()
            queue.extend(dep_path)

            # while queue is not empty
            while len(queue) > 0:
                # take one path
                item = queue.popleft()

                # extend our dependencies and queue by dependencies of this taken path form queue
                for item_path in self.path_to_path[item]:
                    if item_path not in self.full_path_to_path[path]:
                        self.full_path_to_path[path].add(item_path)
                        queue.append(item_path)



import logging
import os
from enum import Enum
from pathlib import Path
from typing import Any, MutableMapping, Optional

from airflow.models import Variable
from airflow.utils.context import Context
from cachetools import TTLCache, cached

from dbt_af.common.constants import (
    DBT_CLI_COMMAND_EXTRA_FLAGS,
    DBT_CLI_COMMAND_EXTRA_OPTIONS,
    OTHER_DBT_CLI_OPTIONS,
    OTHER_DBT_CLI_OPTIONS_DEFAULT,
)
from dbt_af.conf import Config


class TestTag(Enum):
    """DBT tag that specifies a test scheduling strategy in an Airflow DAG."""

    small = '@small'
    medium = '@medium'
    large = '@large'


def get_variable(var_name):
    if var_name in os.environ:
        return os.environ.get(var_name)

    return Variable.get(var_name, f'{{{{ var.value.{var_name} }}}}')


@cached(cache=TTLCache(maxsize=1024, ttl=60 * 5))
def init_environment(config: Config):
    dbt_env = {
        'DBT_TARGET_PATH': str(config.dbt_project.dbt_target_path),
        'DBT_LOG_PATH': str(config.dbt_project.dbt_log_path),
        'DBT_SCHEMA': config.dbt_project.dbt_schema,
        'DBT_PROJECT_DIR': str(config.dbt_project.dbt_project_path),
        'DBT_DEPS_DIR': os.path.join(config.dbt_project.dbt_project_path, 'dbt_packages'),
        'DBT_PROFILES_DIR': str(config.dbt_project.dbt_profiles_path),
        **config.dbt_project.additional_dbt_env,
    }

    return dbt_env


def find_latest_log_file(context: 'Context', log_dir: Path) -> Optional[str]:
    log_pattern = f'dag_id={context["dag"].dag_id}/run_id={context["run_id"]}/task_id={context["task"].task_id}/*.log'
    try:
        return str(max(log_dir.glob(log_pattern)))
    except TypeError:
        logging.error(f'No log file found for dag_id={context["dag"].dag_id}, task_id={context["task"].task_id}')
    except Exception as ex:
        logging.error(f'Something went wrong while searching log file, {ex}')

    return None


def build_dbt_run_model_bash_extra_options(params: MutableMapping[str, Any]) -> tuple[dict[str, str], set[str]]:
    """
    Build extra options for dbt run model command.
    Used only for dbt_run_model DAG.

    Returns a string with extra options for bash command like ` --option_name option_value --flag_name`.
    """
    bash_options = {}
    bash_flags = set()
    for option_name in params.get(DBT_CLI_COMMAND_EXTRA_OPTIONS, []):
        option_value = params.get(option_name)
        if option_value:
            bash_options[f'--{option_name}'] = option_value
    for flag_name in params.get(DBT_CLI_COMMAND_EXTRA_FLAGS, []):
        flag_value = params.get(flag_name)
        if flag_value:
            bash_flags.add(f'--{flag_name}')
    if params.get(OTHER_DBT_CLI_OPTIONS, '') != OTHER_DBT_CLI_OPTIONS_DEFAULT:
        for option_name, option_value in params.get(OTHER_DBT_CLI_OPTIONS, {}).items():
            option_name = option_name if option_name.startswith('--') else f'--{option_name}'
            bash_options[option_name] = option_value

    return bash_options, bash_flags



from importlib.util import find_spec

from dbt_af.integrations.tableau.extracts import tableau_extracts_refresh


def is_tableau_installed():
    return find_spec('tableauserverclient') is not None


__all__ = [
    'is_tableau_installed',
    'tableau_extracts_refresh',
]



from dbt_af.parser.dbt_node_model import DbtNode


def test_all_dbt_tests_have_unique_name(manifest):
    unique_tests_names = set()
    for raw_node in manifest['nodes'].values():
        node = DbtNode(**raw_node)
        if node.is_test():
            test_name = f'{node.name}'

            assert test_name not in unique_tests_names
            unique_tests_names.add(test_name)


def test_dbt_node_name_is_same_as_path(manifest):
    for raw_node in manifest['nodes'].values():
        node = DbtNode(**raw_node)
        if node.is_model():
            real_node_name = node.fqn[-1]
            expected_node_name = f'{node.fqn[1]}.{node.fqn[2]}.{node.name.split(".")[-1]}'
            assert real_node_name == expected_node_name


def test_all_medium_tests_have_parents(dbt_af_graph):
    for node in dbt_af_graph.dbt_nodes:
        if node.is_medium_test():
            dbt_af_graph._find_parent_node_for_test(node)



import os
from typing import Any, Dict, List, Optional

try:
    import pydantic.v1 as pydantic
except ModuleNotFoundError:
    import pydantic


class _FreshnessAfterModel(pydantic.BaseModel):
    count: Optional[int]
    period: Optional[str]

    def empty(self) -> bool:
        return self.count is None and self.period is None


class DbtSourceFreshness(pydantic.BaseModel):
    warn_after: _FreshnessAfterModel
    error_after: _FreshnessAfterModel
    filter: Optional[str] = pydantic.Field(default=None)


class DbtSourceConfig(pydantic.BaseModel):
    enabled: bool


class DbtSource(pydantic.BaseModel):
    database: Optional[str]
    node_schema: Optional[str] = pydantic.Field(..., alias='schema')
    name: str
    resource_type: str
    package_name: str
    path: str
    original_file_path: str
    unique_id: str
    fqn: List[str]
    source_name: str
    source_description: str
    loader: str
    identifier: str
    quoting: Dict[str, Any]
    loaded_at_field: str = pydantic.Field(default=None)
    freshness: DbtSourceFreshness
    external: Optional[Dict[str, Any]] = pydantic.Field(default=None)
    description: str
    columns: Dict[str, Any]
    meta: Dict[str, Any]
    source_meta: Dict[str, Any]
    tags: List[str]
    config: DbtSourceConfig
    patch_path: Optional[str] = pydantic.Field(default=None)
    unrendered_config: Dict[str, Any]
    relation_name: str
    created_at: float

    def __hash__(self) -> int:
        return hash(self.unique_id)

    def is_at_etl_service(self, etl_service_name) -> bool:
        return etl_service_name in self.original_file_path

    @property
    def original_file_path_dirname(self):
        return os.path.dirname(self.original_file_path)

    @property
    def domain(self) -> str:
        return self.fqn[1]

    def need_to_check_freshness(self) -> bool:
        if not self.config.enabled:
            return False
        if self.freshness.warn_after.empty() and self.freshness.error_after.empty():
            return False

        return True



from airflow.utils.task_group import TaskGroup

from dbt_af.builder.dag_components import DagComponent
from dbt_af.builder.dbt_af_builder import DomainDag
from dbt_af.operators.macros import DbtMaintenanceOperatorFactory
from dbt_af.parser.dbt_node_model import DbtModelMaintenanceType, DbtNode, DbtNodeConfig


class MaintenanceDagComponent(DagComponent):
    def __init__(
        self,
        domain_dag: DomainDag,
        maintenance_type: DbtModelMaintenanceType,
        node_config: 'DbtNodeConfig',
    ):
        name = f'{maintenance_type.value}__{domain_dag.dag_name}'
        super().__init__(name=name, domain_dag=domain_dag, node_config=node_config)
        self.maintenance_type = maintenance_type

        self._model_names: set[DbtNode] = set()

    def __hash__(self):
        return hash(self.safe_name)

    def add_model(self, model_name: DbtNode):
        self._model_names.add(model_name)

    def init_af(self):
        self.af_component = TaskGroup(self.safe_name, dag=self.domain_dag.af_dag)

        for model_name in self._model_names:
            DbtMaintenanceOperatorFactory.create(
                model_name=model_name.resource_name,
                schedule_tag=self.domain_dag.schedule,
                maintenance_type=self.maintenance_type,
                task_group=self.af_component,
                af_dag=self.domain_dag.af_dag,
                maintenance_config=model_name.config.maintenance,
                dbt_af_config=self.domain_dag.config,
            )



import datetime

DBT_SENSOR_POOL = 'dbt_sensor_pool'
DBT_COMPILE_POOL = 'dbt_compile_pool'

# dbt_run_model DAG's params
DBT_MODEL_DAG_PARAM = 'dbt_select_model'
OTHER_DBT_CLI_OPTIONS = 'other_dbt_cli_options'
OTHER_DBT_CLI_OPTIONS_DEFAULT = {'--option': 'value'}
DBT_CLI_COMMAND_EXTRA_OPTIONS = '_dbt_cli_command_extra_options'
DBT_CLI_COMMAND_EXTRA_FLAGS = '_dbt_cli_command_extra_flags'

DEFAULT_DAG_ARGS = {'owner': 'airflow', 'retries': 1, 'retry_delay': datetime.timedelta(minutes=1)}

# tag for DBT dags in airflow that have regular schedule (not @manual)
FRONTIER_TAG = 'frontier'
BACKFILL_TAG = 'backfill'
MAINTENANCE_TAG = 'dbt_maintenance'
LARGE_TESTS_TAG = 'dbt_large_tests'

DOMAIN_DAG_START_DATE_FMT = 'YYYY-MM-DDTHH:mm:ss'

# k8s specific constants
AZ_MI_BINDING_LABEL_NAME = 'aadpodidbinding'



import datetime as dt
import json
from typing import Optional

import yaml
from airflow.models.dag import DAG
from airflow.models.param import Param

from dbt_af.builder.dbt_af_builder import BackfillDomainDag, DbtAfGraph, get_domain_dag_start_date
from dbt_af.common.af_callbacks import collect_af_custom_callbacks
from dbt_af.common.constants import (
    DBT_CLI_COMMAND_EXTRA_FLAGS,
    DBT_CLI_COMMAND_EXTRA_OPTIONS,
    DBT_MODEL_DAG_PARAM,
    DEFAULT_DAG_ARGS,
    OTHER_DBT_CLI_OPTIONS,
    OTHER_DBT_CLI_OPTIONS_DEFAULT,
)
from dbt_af.conf import Config
from dbt_af.operators.run import DbtRun


def dbt_main_dags(graph: DbtAfGraph) -> dict[str, DAG]:
    af_dags = {}

    dag_callbacks, task_callbacks = collect_af_custom_callbacks(graph.config)
    domains = {node.domain_dag for node in graph.nodes}

    for domain_dag in domains:
        dag = DAG(
            domain_dag.dag_name,
            start_date=get_domain_dag_start_date(graph, domain_dag),
            description=graph.config.af_dag_description,
            schedule=domain_dag.schedule.af_repr(),
            catchup=domain_dag.catchup if not graph.config.dry_run else False,
            default_args=DEFAULT_DAG_ARGS,
            max_active_runs=graph.config.max_active_dag_runs,
            render_template_as_native_obj=False,
            tags=['dbt'] + domain_dag.tags,
            **dag_callbacks,
        )
        domain_dag.af_dag = dag
        af_dags[domain_dag.dag_name] = dag

        if isinstance(domain_dag, BackfillDomainDag):
            domain_dag.wrap_dag_with_endpoints()

    for node in graph.nodes:
        node.domain_dag.af_dag = af_dags[node.domain_dag.dag_name]

    for node in graph.nodes:
        node.add_af_callbacks(task_callbacks)
        if node.af_component is None:
            node.init_af()

    for node in graph.nodes:
        if isinstance(node.domain_dag, BackfillDomainDag):
            start_task = node.domain_dag.start_endpoint
            if len(node.af_component.upstream_task_ids) == 0:
                start_task >> node.af_component

    return af_dags


def dbt_run_model_dag(config: Config) -> dict[str, DAG]:
    dbt_project_name = config.dbt_project.dbt_project_name
    dag_name = f'{dbt_project_name}_dbt_run_model'

    dag_callbacks, task_callbacks = collect_af_custom_callbacks(config)
    dag = DAG(
        dag_name,
        start_date=config.dag_start_date,
        description=config.af_dag_description,
        schedule_interval=None,
        catchup=False,
        default_args=DEFAULT_DAG_ARGS,
        max_active_runs=config.max_active_dag_runs,
        tags=[dbt_project_name, 'dbt', 'system'],
        params={
            DBT_MODEL_DAG_PARAM: Param(
                '',
                type='string',
                title='Models selector',
                description_md=(
                    'Refer to the [dbt documentation](https://docs.getdbt.com/reference/node-selection/syntax) '
                    'for detailed information on how to specify the models selector.'
                ),
            ),
            'start_dttm': Param(
                f'{dt.date.today()-dt.timedelta(days=1)}T{dt.time(hour=0, minute=0, second=0)}',
                type='string',
                format='date-time',
                title='Interval start datetime',
                description_md='Could be accessible in dbt model using `{{ var("start_dttm") }}`',
            ),
            'end_dttm': Param(
                f'{dt.date.today()}T{dt.time(hour=0, minute=0, second=0)}',
                type='string',
                format='date-time',
                title='Interval end datetime',
                description_md='Could be accessible in dbt model using `{{ var("end_dttm") }}`',
            ),
            'target': Param(
                '',
                description_md=(
                    'Optional override of target environment (used as `--target` option). '
                    'Available target environments can be found in `profiles.yml`.'
                ),
                type=['null', 'string'],
            ),
            'full-refresh': Param(
                False,
                type='boolean',
                title='Full refresh',
                description_md=(
                    'If specified, dbt will drop incremental models and fully-recalculate the incremental '
                    'table from the model definition.'
                ),
            ),
            OTHER_DBT_CLI_OPTIONS: Param(
                OTHER_DBT_CLI_OPTIONS_DEFAULT,
                type=['object', 'null'],
                title='Extra arguments',
                description_md=(
                    'Extra arguments to pass to dbt run command. See `dbt run --help` for more details. '
                    'Will be passed to the command as `--option1 value1 --option2 value2`'
                ),
                section='Extra arguments',
            ),
            DBT_CLI_COMMAND_EXTRA_OPTIONS: Param(
                ['target'],
                const=['target'],
            ),
            DBT_CLI_COMMAND_EXTRA_FLAGS: Param(
                ['full-refresh'],
                const=['full-refresh'],
            ),
        },
        **dag_callbacks,
    )

    target_environment = config.dbt_default_targets.default_target
    DbtRun(
        task_id='dbt_model',
        model_name=None,
        dag=dag,
        target_environment=target_environment,
        dbt_af_config=config,
        **task_callbacks,
    )

    return {dag_name: dag}


def _compile_dbt_dags(
    manifest_content: dict,
    profiles: dict,
    project_profile_name: str,
    config: Config,
    etl_service_name: Optional[str] = None,
) -> dict[str, DAG]:
    dags = {}

    graph = DbtAfGraph.from_manifest(
        manifest_content, profiles, project_profile_name, etl_service_name=etl_service_name, config=config
    )

    dags.update(dbt_main_dags(graph))
    if config.include_single_model_manual_dag:
        dags.update(dbt_run_model_dag(config=config))

    return dags


def compile_dbt_af_dags(manifest_path: str, config: Config, etl_service_name: Optional[str] = None) -> dict[str, DAG]:
    """
    Compiles airflow DAGs from manifest according to provided dbt-af config.
    It's possible to use different etl service names for different model groups in one dbt project.
    """

    with open(manifest_path) as fin:
        manifest = json.load(fin)

    with open(config.dbt_project.dbt_profiles_path / 'profiles.yml') as fin:
        profiles = yaml.safe_load(fin)

    with open(config.dbt_project.dbt_project_path / 'dbt_project.yml') as fin:
        dbt_project_profile_name = yaml.safe_load(fin)['profile']

    return _compile_dbt_dags(
        manifest, profiles, dbt_project_profile_name, etl_service_name=etl_service_name, config=config
    )



from collections import defaultdict
from typing import Generator, Optional

from airflow.operators.empty import EmptyOperator
from airflow.utils.task_group import TaskGroup

from dbt_af.builder.domain_dag import DomainDag
from dbt_af.builder.task_dependencies import DagDelayedDependencyRegistry
from dbt_af.common.scheduling import EScheduleTag
from dbt_af.operators.branch import DbtBranchOperator, create_decision_path_function
from dbt_af.operators.kubernetes_pod import DbtKubernetesPodOperator
from dbt_af.operators.run import DbtRun, DbtSeed, DbtSnapshot, DbtTest
from dbt_af.operators.sensors import AfExecutionDateFn, DbtExternalSensor, DbtSourceFreshnessSensor
from dbt_af.operators.supplemental import TableauExtractsRefreshOperator
from dbt_af.operators.venv import DbtPythonVenvOperator
from dbt_af.parser.dbt_node_model import DbtNode, DbtNodeConfig
from dbt_af.parser.dbt_profiles import KubernetesTarget, VenvTarget
from dbt_af.parser.dbt_source_model import DbtSource


class DagComponent:
    add_external_dependencies = True

    def __init__(self, name: str, domain_dag: DomainDag, node_config: DbtNodeConfig):
        self.name = name
        self.domain_dag = domain_dag
        self.node_config = node_config

        # dbt-af components
        self._depends_on: set[DagComponent] = set()
        self._depends_on_sources: set[DbtSource] = set()
        self._domains_dependencies: dict[DomainDag, set[DagComponent]] = defaultdict(set)
        self.delayed_deps_registry = DagDelayedDependencyRegistry()
        self._small_tests: set[str] = set()

        # airflow components
        self.af_component: Optional[DbtRun | DbtKubernetesPodOperator | TaskGroup] = None
        self.model_task: Optional[DbtRun | DbtKubernetesPodOperator] = None
        self.task_group: Optional[TaskGroup] = None
        self.af_sensor_endpoint: Optional[EmptyOperator | DbtRun | DbtKubernetesPodOperator] = None
        self._af_callbacks: dict[str, list[Optional[callable]]] = {}

    @property
    def depends_on(self) -> list['DagComponent']:
        return list(self._depends_on)

    @property
    def safe_name(self) -> str:
        return self.name.replace('.', '__')

    def add_af_callbacks(self, callbacks: dict[str, list[Optional[callable]]]):
        self._af_callbacks.update(callbacks)

    def add_dependency(self, dep: 'DagComponent'):
        if self.node_config.dependencies[dep.name].skip:
            return

        self._depends_on.add(dep)
        if self.domain_dag.config.model_dependencies.wait_policy.per_domain:
            self._domains_dependencies[dep.domain_dag].add(dep)

    def add_source_dependency(self, dep: DbtSource):
        self._depends_on_sources.add(dep)

    def add_small_test(self, resource_name: str):
        self._small_tests.add(resource_name)

    def _create_opt_brancher(self, delayed_deps: DagDelayedDependencyRegistry) -> Optional[DbtBranchOperator]:
        """
        Create a brancher task to decide if the model should be run or not based on the enable_from_dttm and
        disable_from_dttm parameters
        """
        if not self.node_config.enable_from_dttm and not self.node_config.disable_from_dttm:
            return None

        brancher = DbtBranchOperator(
            task_id=self.safe_name,
            task_group=self.task_group,
            python_callable=create_decision_path_function(self.node_config, self.safe_name),
            dag=self.domain_dag.af_dag,
        )
        delayed_deps(brancher) >> delayed_deps(self.model_task)

        return brancher

    def _ext_dep_waits_generator(
        self,
        dep: 'DagComponent',
        task_group: TaskGroup,
    ) -> Generator[DbtExternalSensor, None, None]:
        execution_date_fns = AfExecutionDateFn(
            upstream_schedule_tag=dep.domain_dag.schedule,
            downstream_schedule_tag=self.domain_dag.schedule,
            wait_policy=self.node_config.dependencies[dep.name].wait_policy,
        ).get_execution_dates()

        for i, execution_date_fn in enumerate(execution_date_fns):
            # airflow task_id for statsd must be less than 250 chars.
            # it's not necessary to have a long name for the only one external dependency wait
            _suffix = f'__{i}' if len(execution_date_fns) > 1 else ''
            wait = DbtExternalSensor(
                dbt_af_config=self.domain_dag.config,
                task_id=f'wait__{dep.safe_name}{_suffix}',
                task_group=task_group,
                external_dag_id=dep.domain_dag.af_dag.dag_id,
                external_task_id=dep.af_sensor_endpoint.task_id,
                execution_date_fn=execution_date_fn,
                dep_schedule=dep.domain_dag.schedule,
                dag=self.domain_dag.af_dag,
            )
            yield wait

    def _is_external_dep_valid(self, dep: 'DagComponent') -> bool:
        return (
            dep.domain_dag != self.domain_dag
            and self.add_external_dependencies
            and dep.domain_dag.schedule != EScheduleTag.manual()
            and self.domain_dag.schedule != EScheduleTag.manual()
        )

    def _init_dependencies_per_domain_af(self, delayed_deps: DagDelayedDependencyRegistry):
        for dep_domain_dag, deps in self._domains_dependencies.items():
            for dep in deps:
                if not self._is_external_dep_valid(dep):
                    continue

                deps_registry = self.domain_dag.registered_domains_dependencies[dep_domain_dag]
                if not deps_registry.is_registered(dep):
                    if not deps_registry.task_group:
                        deps_registry.task_group = TaskGroup(
                            group_id=f'{dep_domain_dag.dag_name}__dependencies__group',
                            dag=self.domain_dag.af_dag,
                        )
                    for wait in self._ext_dep_waits_generator(dep, deps_registry.task_group):
                        deps_registry.add_dependency(dep, wait)

                for wait_task in deps_registry.get_dependency_wait_task(dep):
                    delayed_deps(wait_task) >> delayed_deps(self.model_task)

    def _init_dependencies_per_task_af(
        self,
        delayed_deps: DagDelayedDependencyRegistry,
        brancher: Optional[DbtBranchOperator],
    ):
        for dep in self._depends_on:
            if not self._is_external_dep_valid(dep):
                continue

            for wait in self._ext_dep_waits_generator(dep, self.af_component):
                delayed_deps(brancher) >> delayed_deps(wait)
                delayed_deps(wait) >> delayed_deps(self.model_task)

    def _init_dependencies_af(self, delayed_deps: DagDelayedDependencyRegistry):
        """
        Create dependencies between the model and its upstreams. If the model has external dependencies, create
        external sensor tasks to wait for the upstreams to finish
        """
        brancher = self._create_opt_brancher(delayed_deps=delayed_deps)

        for dep in self._depends_on:
            if dep.af_component is None:
                dep.init_af()

            if dep.domain_dag == self.domain_dag:
                delayed_deps(dep.model_task) >> delayed_deps(self.model_task)
                delayed_deps(dep.af_component) >> delayed_deps(self.af_component)
                delayed_deps(dep.model_task) >> delayed_deps(brancher)

        if self.domain_dag.config.model_dependencies.wait_policy.per_domain:
            self._init_dependencies_per_domain_af(delayed_deps)
        elif self.domain_dag.config.model_dependencies.wait_policy.per_task:
            self._init_dependencies_per_task_af(delayed_deps, brancher)
        else:
            raise ValueError(
                f'Unknown wait policy (or all policies are turned off): '
                f'{self.domain_dag.config.model_dependencies.wait_policy}'
            )

    def _get_ext_deps(self) -> list['DagComponent']:
        return [dep for dep in self._depends_on if self._is_external_dep_valid(dep)]

    def _get_source_deps_with_freshness_check(self) -> list[DbtSource]:
        return [dep for dep in self._depends_on_sources if dep.need_to_check_freshness()]

    def _create_task_group(self) -> Optional[TaskGroup]:
        """
        Create a task group for the model if it has external dependencies or small tests. If waits for all external
        dependencies are built per domain, a task group is not needed
        """
        if (
            not self._small_tests
            and (not self._get_ext_deps() or self.domain_dag.config.model_dependencies.wait_policy.per_domain)
            and not self._get_source_deps_with_freshness_check()
            and not self.node_config.enable_from_dttm
            and not self.node_config.disable_from_dttm
            and not self.node_config.tableau_refresh_tasks
        ):
            return None

        return TaskGroup(f'{self.safe_name}__group', dag=self.domain_dag.af_dag)

    def init_af(self):
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(f'{self.name}@{self.domain_dag.dag_name}')

    def __eq__(self, other) -> bool:
        if isinstance(other, DagComponent):
            return self.name == other.name and self.domain_dag == other.domain_dag
        raise TypeError(f'Cannot compare {self} with {other}')

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.name}, {self.domain_dag})'


class DagModel(DagComponent):
    runner_class = DbtRun
    add_external_dependencies = True
    overlap = True
    is_dataset_enable = True

    def __init__(self, dbt_node: DbtNode, domain_dag: DomainDag):
        super().__init__(dbt_node.resource_name, domain_dag, node_config=dbt_node.config)

        self.dbt_node = dbt_node
        self.target_environment = self.dbt_node.target_environment(domain_dag.config.dbt_default_targets)
        self.max_active_tis_per_dag = self.dbt_node.get_airflow_parallelism()

    def _create_dbt_runner_task(self) -> DbtRun:
        return self.runner_class(
            task_id=self.safe_name,
            model_name=self.name,
            is_dataset_enable=self.is_dataset_enable,
            dag=self.domain_dag.af_dag,
            task_group=self.task_group,
            schedule_tag=self.domain_dag.schedule,
            overlap=self.overlap,
            max_active_tis_per_dag=self.max_active_tis_per_dag,
            model_type=self.dbt_node.model_type,
            target_environment=self.target_environment,
            dbt_af_config=self.domain_dag.config,
            env=self.dbt_node.config.env,
            **self._af_callbacks,
        )

    def _create_k8s_runner_task(self) -> DbtKubernetesPodOperator:
        """
        Create a k8s operator to run the dbt model not in Databricks but in a k8s pod
        It's used only in rare cases when model requires a lot of resources and/or data processing on the same node
        """
        return DbtKubernetesPodOperator(
            task_id=self.safe_name,
            dbt_model_name=self.name,
            dbt_model_path=self.dbt_node.path,
            task_group=self.task_group,
            dag=self.domain_dag.af_dag,
            target_details=self.dbt_node.target_details,
            dbt_af_config=self.domain_dag.config,
            env=self.dbt_node.config.env,
        )

    def _create_venv_runner_task(self) -> DbtPythonVenvOperator:
        return DbtPythonVenvOperator(
            task_id=self.safe_name,
            task_group=self.task_group,
            dag=self.domain_dag.af_dag,
            dbt_model_path=self.dbt_node.path,
            target_details=self.dbt_node.target_details,
            dbt_af_config=self.domain_dag.config,
            env=self.dbt_node.config.env,
        )

    def _create_runner_task(self) -> DbtRun | DbtKubernetesPodOperator | DbtPythonVenvOperator:
        if isinstance(self.dbt_node.target_details, KubernetesTarget):
            return self._create_k8s_runner_task()
        if isinstance(self.dbt_node.target_details, VenvTarget):
            return self._create_venv_runner_task()
        return self._create_dbt_runner_task()

    def _init_small_tests_af(self, delayed_deps: DagDelayedDependencyRegistry) -> Optional[EmptyOperator]:
        """
        Create small tests for the model if it has any. If there are any tests, they will be run after the model and
        after all tests are finished, the empty endpoint task will be created
        """
        if not self._small_tests:
            return None

        endpoint_task = EmptyOperator(
            task_id=f'{self.safe_name}__end',
            task_group=self.task_group,
            dag=self.domain_dag.af_dag,
        )
        for test in self._small_tests:
            test_task = DbtTest(
                task_id=test.replace('.', '__'),
                model_name=test,
                dag=self.domain_dag.af_dag,
                task_group=self.task_group,
                schedule_tag=self.domain_dag.schedule,
                dbt_af_config=self.domain_dag.config,
            )
            delayed_deps(self.model_task) >> delayed_deps(test_task)
            delayed_deps(test_task) >> delayed_deps(endpoint_task)

        return endpoint_task

    def _init_source_dependencies_af(self, delayed_deps: DagDelayedDependencyRegistry):
        for source_dep in self._depends_on_sources:
            if source_dep.need_to_check_freshness():
                source_wait = DbtSourceFreshnessSensor(
                    task_id=f'wait_freshness__{source_dep.name}__for__{self.safe_name}',
                    task_group=self.af_component,
                    dag=self.domain_dag.af_dag,
                    env=self.model_task.env if hasattr(self.model_task, 'env') else {},
                    source_name=source_dep.source_name,
                    source_identifier=source_dep.identifier,
                    dbt_af_config=self.domain_dag.config,
                )

                delayed_deps(source_wait) >> delayed_deps(self.model_task)

    def _init_supplemental_dependencies_af(self, delayed_deps: DagDelayedDependencyRegistry):
        if self.dbt_node.config.tableau_refresh_tasks:
            tableau_refresh_task = TableauExtractsRefreshOperator(
                task_id=f'tableau_refresh__{self.safe_name}',
                task_group=self.task_group,
                dag=self.domain_dag.af_dag,
                tableau_refresh_tasks=self.dbt_node.config.tableau_refresh_tasks,
                dbt_af_config=self.domain_dag.config,
            )
            delayed_deps(self.model_task) >> delayed_deps(tableau_refresh_task)

    def init_af(self):
        """
        Initialize all Airflow components for the dbt-model and it's dependencies
        """
        if self.domain_dag.af_dag is None:
            raise ValueError(f'{self!r}: dag not set')

        with self.delayed_deps_registry as delayed_deps:
            self.task_group = self._create_task_group()
            self.model_task = self._create_runner_task()
            endpoint_task = self._init_small_tests_af(delayed_deps)

            self.af_component = self.task_group or self.model_task
            self.af_sensor_endpoint = endpoint_task or self.model_task

            self._init_dependencies_af(delayed_deps)
            self._init_source_dependencies_af(delayed_deps)
            self._init_supplemental_dependencies_af(delayed_deps)


class DagSnapshot(DagModel):
    runner_class = DbtSnapshot


class DagSeed(DagModel):
    runner_class = DbtSeed


class MediumTests(DagComponent):
    def __init__(self, domain_dag: DomainDag, node_config: DbtNodeConfig):
        name = f'medium_tests__{domain_dag.dag_name}'
        super().__init__(name, domain_dag, node_config=node_config)
        self._tests: set[str] = set()

    @staticmethod
    def get_medium_test_name(node: DbtNode, parent_model: DagModel) -> str:
        return f'{parent_model.safe_name}__{node.resource_name}'

    def add_test(self, node_id: str):
        self._tests.add(node_id)

    def init_af(self):
        with self.delayed_deps_registry as delayed_deps:
            self.af_component = TaskGroup(self.safe_name, dag=self.domain_dag.af_dag)

            for test in self._tests:
                DbtTest(
                    task_id=test.replace('.', '__'),
                    model_name=test,
                    task_group=self.af_component,
                    dag=self.domain_dag.af_dag,
                    schedule_tag=self.domain_dag.schedule,
                    dbt_af_config=self.domain_dag.config,
                )

            for dep in self.depends_on:
                if dep.af_component is None:
                    dep.init_af()

                delayed_deps(dep.af_component) >> delayed_deps(self.af_component)


class LargeTest(DagComponent):
    add_external_dependencies = True

    def __init__(self, name: str, domain_dag: DomainDag, node_config: DbtNodeConfig):
        super().__init__(name, domain_dag, node_config=node_config)

    def init_af(self):
        with self.delayed_deps_registry as delayed_deps:
            self.task_group = self._create_task_group()

            self.model_task = DbtTest(
                task_id=self.safe_name,
                model_name=self.name,
                dag=self.domain_dag.af_dag,
                task_group=self.task_group,
                schedule_tag=self.domain_dag.schedule,
                dbt_af_config=self.domain_dag.config,
            )
            self.af_component = self.task_group or self.model_task

            self._init_dependencies_af(delayed_deps)



import logging
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from pycarlo.core import Client


def _get_resource_id(client: 'Client', metastore_name: str) -> str:
    """
    Get the resource ID by the metastore name

    copy function from https://github.com/mvfolino68/pycarlo-examples/blob/main/dbt/dbt.py#L16
    """

    from pycarlo.core import Query

    query = Query()
    query.get_user().account.warehouses.__fields__('name', 'connection_type', 'uuid')
    warehouses = client(query).get_user.account.warehouses
    for wh in warehouses:
        if wh.name == metastore_name:
            return wh.uuid
    raise ValueError(f'Warehouse {metastore_name} not found. Found: {warehouses}')


def send_dbt_artifacts_to_montecarlo(
    target_path: str,
    log_path: Optional[str],
    model_name: str,
    metastore_name: str,
    project_name: str,
):
    """
    Send dbt artifacts to MonteCarloData.

    NOTE: This function requires default mcd connections (not Gateway Connection)
    """
    from airflow_mcd.hooks import SessionHook
    from pycarlo.core import Client
    from pycarlo.features.dbt import DbtImporter

    logging.info('Sending dbt artifacts to MonteCarlo')

    conn = SessionHook(mcd_session_conn_id=SessionHook.default_conn_name).get_conn()
    mc_client = Client(session=conn)
    resource_id = _get_resource_id(mc_client, metastore_name)

    DbtImporter(mc_client=mc_client).import_run(
        manifest_path=f'{target_path}/manifest.json',
        run_results_path=f'{target_path}/run_results.json',
        logs_path=log_path,
        project_name=project_name,
        job_name=f'dbt_run_{model_name}',
        resource_id=resource_id,
    )
    logging.info('Successfully sent dbt artifacts to MonteCarlo')



from typing import TYPE_CHECKING

from dbt_af.builder.dag_components import DagModel
from dbt_af.operators.run import DbtRun, DbtSnapshot

if TYPE_CHECKING:
    from dbt_af.builder.dbt_af_builder import DomainDag
    from dbt_af.parser.dbt_node_model import DbtNode


class BackfillDagModel(DagModel):
    runner_class = DbtRun
    max_active_tis_per_dag = 1
    add_external_dependencies = False
    overlap = False
    is_dataset_enable = False

    def __init__(self, dbt_node: 'DbtNode', domain_dag: 'DomainDag'):
        super().__init__(dbt_node, domain_dag)
        self.target_environment = (
            self.dbt_node.config.bf_cluster or domain_dag.config.dbt_default_targets.default_backfill_target
        )

    @property
    def safe_name(self) -> str:
        return f'{super().safe_name}__bf'

    def __str__(self) -> str:
        return f'{self.name}__bf'

    def __hash__(self) -> int:
        return hash(f'{self.name}@bf@{self.domain_dag.dag_name}')

    def _get_ext_deps(self) -> list:
        return []


class BackfillDagSnapshot(BackfillDagModel):
    runner_class = DbtSnapshot



from collections import defaultdict
from enum import Enum
from typing import Optional

from airflow.operators.empty import EmptyOperator
from airflow.operators.python import BranchPythonOperator

from dbt_af.builder.task_dependencies import RegistryDomainDependencies
from dbt_af.common import constants
from dbt_af.common.scheduling import BaseScheduleTag, EScheduleTag
from dbt_af.conf import Config


class DomainDagType(Enum):
    SCHEDULED = 'scheduled'
    BACKFILL = 'backfill'
    MAINTENANCE = 'maintenance'
    LARGE_TESTS = 'large_tests'


class DomainDag:
    def __init__(
        self,
        domain_name: str,
        schedule: BaseScheduleTag = None,
        config: Optional[Config] = None,
        additional_tags: Optional[list[str]] = None,
        catchup: bool = True,
    ):
        self.domain_name = domain_name
        self._schedule = schedule or EScheduleTag.daily()
        self.config: Config = config or Config()

        self.additional_tags: list[str] = additional_tags or []
        self.catchup = catchup

        self.registered_domains_dependencies: dict[DomainDag, RegistryDomainDependencies] = defaultdict(
            RegistryDomainDependencies
        )

        self.af_dag = None

    @property
    def _base_tags(self) -> Optional[list[str]]:
        return None

    @property
    def dag_name(self) -> str:
        return f'{self.domain_name}{self._schedule}'.replace('@', '__')

    @property
    def schedule(self) -> BaseScheduleTag:
        return self._schedule

    @schedule.setter
    def schedule(self, value: BaseScheduleTag):
        self._schedule = value

    @property
    def tags(self) -> list[str]:
        pure_domain_name = self.domain_name.split('__')[0]
        merged_tags = (self._base_tags or []) + [pure_domain_name, self.schedule.safe_name] + self.additional_tags
        if self.schedule != EScheduleTag.manual() and constants.BACKFILL_TAG not in merged_tags:
            merged_tags.append(constants.FRONTIER_TAG)

        return merged_tags

    def __hash__(self) -> int:
        return hash(self.dag_name)

    def __eq__(self, other) -> bool:
        if isinstance(other, DomainDag):
            return self.dag_name == other.dag_name
        raise TypeError(f'Cannot compare {self.__class__.__name__} with {other.__class__.__name__}')

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.domain_name}, {self._schedule})'


class BackfillDomainDag(DomainDag):
    def __init__(
        self,
        domain_name: str,
        config: Optional[Config] = None,
        additional_tags: list[str] = None,
        catchup: bool = True,
    ):
        super().__init__(domain_name, self.schedule, config, additional_tags, catchup)

        self.start_endpoint: Optional[EmptyOperator] = None

    @property
    def _base_tags(self) -> Optional[list[str]]:
        return [constants.BACKFILL_TAG]

    @property
    def dag_name(self) -> str:
        return f'{self.domain_name}__backfill'.replace('@', '__')

    @property
    def schedule(self) -> BaseScheduleTag:
        return EScheduleTag.daily()

    def wrap_dag_with_endpoints(self):
        """
        Each backfill dag should have start with branch operator and end with empty operator.
        For the first scheduled run branch operator will return 'do_nothing' task and all dbt tasks will be skipped.
        If airflow dag is triggered again after scheduled run, it will trigger all downstream dbt tasks.
        """
        self.start_endpoint = EmptyOperator(task_id='start_work', dag=self.af_dag)
        do_nothing = EmptyOperator(task_id='do_nothing', dag=self.af_dag)

        def decide_which_path(**kwargs):
            if kwargs['task_instance'].try_number > 1:
                return 'start_work'
            return 'do_nothing'

        brancher = BranchPythonOperator(task_id='branch', python_callable=decide_which_path, dag=self.af_dag)
        brancher >> [self.start_endpoint, do_nothing]


class MaintenanceDomainDag(DomainDag):
    def __init__(
        self,
        domain_name: str,
        config: Optional[Config] = None,
        additional_tags: list[str] = None,
        catchup: bool = False,
    ):
        super().__init__(domain_name, self.schedule, config, additional_tags, catchup)

    @property
    def _base_tags(self) -> Optional[list[str]]:
        return [constants.MAINTENANCE_TAG]

    @property
    def dag_name(self) -> str:
        return f'{self.domain_name}__maintenance'.replace('@', '__')

    @property
    def schedule(self) -> BaseScheduleTag:
        return EScheduleTag.daily()


class LargeTestsDomainDag(DomainDag):
    def __init__(
        self,
        domain_name: str,
        config: Optional[Config] = None,
        additional_tags: list[str] = None,
        catchup: bool = False,
    ):
        super().__init__(domain_name, self.schedule, config, additional_tags, catchup)

    @property
    def _base_tags(self) -> Optional[list[str]]:
        return [constants.LARGE_TESTS_TAG]

    @property
    def dag_name(self) -> str:
        return f'{self.domain_name}__large_tests{self.schedule}'.replace('@', '__')

    @property
    def schedule(self) -> BaseScheduleTag:
        return EScheduleTag.daily()


class DomainDagFactory:
    @staticmethod
    def create(dag_type: DomainDagType, domain_name: str, schedule: BaseScheduleTag, config: Config) -> DomainDag:
        if dag_type == DomainDagType.SCHEDULED:
            return DomainDag(domain_name, schedule, config)
        if dag_type == DomainDagType.BACKFILL:
            return BackfillDomainDag(domain_name, config)
        if dag_type == DomainDagType.MAINTENANCE:
            return MaintenanceDomainDag(domain_name, config)
        if dag_type == DomainDagType.LARGE_TESTS:
            return LargeTestsDomainDag(domain_name, config)

        raise TypeError(f'Unknown dag type: {dag_type}')



from functools import partial
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowSkipException
from cachetools import TTLCache, cachedmethod
from cachetools.keys import hashkey

from dbt_af.integrations.tableau.auth import get_tableau_auth_object
from dbt_af.integrations.tableau.exceptions import (
    FailedTableauRefreshTasksException,
    UnknownTableauResourceTypeException,
)
from dbt_af.parser.dbt_node_model import TableauRefreshResourceType, TableauRefreshTaskConfig

if TYPE_CHECKING:
    import tableauserverclient

    from dbt_af.conf import Config

EXTRACT_OPERATION_ALREADY_QUEUED = '409093'


class _TableauExtractsRegistry:
    def __init__(self, server: 'tableauserverclient.Server'):
        self.server = server

        self._cache: TTLCache = TTLCache(maxsize=1024, ttl=60 * 15)

    @cachedmethod(lambda self: self._cache, key=partial(hashkey, '_get_workbooks_id_mapping'))
    def _get_workbooks_id_mapping(self) -> dict[tuple[str, str], str]:
        import tableauserverclient as tsc

        workbooks_id_mapping: dict[tuple[str, str], str] = {}
        for wb in tsc.Pager(self.server.workbooks):
            workbooks_id_mapping[(wb.project_name, wb.name)] = wb.id

        return workbooks_id_mapping

    @cachedmethod(lambda self: self._cache, key=partial(hashkey, '_get_datasources_id_mapping'))
    def _get_datasources_id_mapping(self) -> dict[tuple[str, str], str]:
        import tableauserverclient as tsc

        datasources_id_mapping: dict[tuple[str, str], str] = {}
        for ds in tsc.Pager(self.server.datasources):
            datasources_id_mapping[(ds.project_name, ds.name)] = ds.id

        return datasources_id_mapping

    def get_resource_id(self, resource_type: TableauRefreshResourceType, project_name: str, resource_name: str) -> str:
        match resource_type:
            case TableauRefreshResourceType.workbook:
                return self._get_workbooks_id_mapping()[(project_name, resource_name)]
            case TableauRefreshResourceType.datasource:
                return self._get_datasources_id_mapping()[(project_name, resource_name)]
            case _:
                raise UnknownTableauResourceTypeException(f'Unknown resource type: {resource_type}')


def tableau_extracts_refresh(tableau_refresh_tasks: 'list[TableauRefreshTaskConfig]', dbt_af_config: 'Config') -> None:
    import tableauserverclient as tsc

    tableau_auth = get_tableau_auth_object(dbt_af_config)
    tableau_server = tsc.Server(server_address=dbt_af_config.tableau.server_address, use_server_version=True)
    tableau_server.auth.sign_in(tableau_auth)

    extracts_registry = _TableauExtractsRegistry(tableau_server)

    failed_tasks: list[TableauRefreshTaskConfig] = []
    for refresh_task in tableau_refresh_tasks:
        try:
            resource_id = extracts_registry.get_resource_id(
                refresh_task.resource_type,
                refresh_task.project_name,
                refresh_task.resource_name,
            )
            match refresh_task.resource_type:
                case TableauRefreshResourceType.workbook:
                    tableau_server.workbooks.refresh(resource_id)
                case TableauRefreshResourceType.datasource:
                    tableau_server.datasources.refresh(resource_id)
                case _:
                    raise UnknownTableauResourceTypeException(f'Unknown resource type: {refresh_task.resource_type}')
        except tsc.ServerResponseError as sre:
            if sre.code == EXTRACT_OPERATION_ALREADY_QUEUED:
                raise AirflowSkipException(f'{sre.code}: {sre.summary}')
            raise sre
        except UnknownTableauResourceTypeException:
            failed_tasks.append(refresh_task)

    if not failed_tasks:
        return

    raise FailedTableauRefreshTasksException(f'Failed to refresh the following Tableau resources: {failed_tasks}')



from typing import TYPE_CHECKING

from airflow.exceptions import AirflowSkipException
from airflow.operators.python import PythonOperator

from dbt_af.conf import Config
from dbt_af.integrations.tableau import is_tableau_installed, tableau_extracts_refresh

if TYPE_CHECKING:
    from dbt_af.parser.dbt_node_model import TableauRefreshTaskConfig


def _tableau_extracts_refresh_dev(*args, **kwargs) -> None:
    import logging

    logging.info('tableau_extracts_refresh is disabled in dev mode.')
    raise AirflowSkipException('tableau_extracts_refresh is disabled in dev mode.')


class TableauExtractsRefreshOperator(PythonOperator):
    template_fields = tuple()

    def __init__(self, tableau_refresh_tasks: 'list[TableauRefreshTaskConfig]', dbt_af_config: Config, **kwargs):
        if not is_tableau_installed():
            raise ImportError('tableauserverclient is not installed. Please install it to use this operator.')

        super().__init__(
            python_callable=tableau_extracts_refresh if not dbt_af_config.dry_run else _tableau_extracts_refresh_dev,
            op_kwargs={
                'tableau_refresh_tasks': tableau_refresh_tasks,
                'dbt_af_config': dbt_af_config,
            },
            **dbt_af_config.retries_config.supplemental_task_retry_policy.as_dict(),
            **kwargs,
        )



import datetime
from pathlib import Path
from typing import Any, Optional

import attrs
import pendulum


@attrs.define(frozen=True)
class CustomAfCallbacksConfig:
    """
    Config to define callbacks functions for airflow DAGs and tasks

    :param task_on_success_callback: success callback function for task (invoked when the task succeeds)
    :param task_on_failure_callback: failure callback function for task (invoked when the airflow task fails)
    :param task_on_retry_callback: retry callback function for task (invoked when the task is up for retry)
    :param task_on_execute_callback: execute callback function for task (invoked right before task begins executing)
    :param dag_on_failure_callback: failure callback function for DAG (invoked when the airflow task fails)
    :param dag_on_success_callback: success callback function for DAG (invoked when the task succeeds)
    :param dag_sla_miss_callback: sla miss callback function for DAG (invoked when a task misses its defined SLA)
    """

    task_on_success_callback: tuple[callable] = attrs.field(factory=tuple)
    task_on_failure_callback: tuple[callable] = attrs.field(factory=tuple)
    task_on_retry_callback: tuple[callable] = attrs.field(factory=tuple)
    task_on_execute_callback: tuple[callable] = attrs.field(factory=tuple)

    dag_on_failure_callback: tuple[callable] = attrs.field(factory=tuple)
    dag_on_success_callback: tuple[callable] = attrs.field(factory=tuple)
    dag_sla_miss_callback: tuple[callable] = attrs.field(factory=tuple)


@attrs.define(frozen=True)
class DependencyWaitPolicy:
    """
    Policy to build waits for models' dependencies.
    Default policy is to build waits per domain because it produces fewer tasks in airflow dag. It's possible to
    choose both policies at the same time, but only per domain policy will be used.

    :param per_domain: whether to build waits for models' dependencies per domain; if it's set then all waits will be
        collected in one task group per upstream domain
    :param per_task: whether to build waits for models' dependencies per task; if it's set then all waits will be
        put in the same task group with downstream model
    """

    per_domain: bool = attrs.field(default=True)
    per_task: bool = attrs.field(default=False)

    def __attrs_post_init__(self):
        if not self.per_domain and not self.per_task:
            raise ValueError('At least one of policy must be set')


@attrs.define(frozen=True)
class ModelDependenciesSection:
    """
    Section to parametrize how model dependencies are handled.

    :param wait_policy: policy to build waits for models' dependencies
    """

    wait_policy: DependencyWaitPolicy = attrs.field(factory=DependencyWaitPolicy)


@attrs.define(frozen=True)
class MCDIntegrationConfig:
    """
    Config for Monte Carlo Data integration.

    :param callbacks_enabled: whether to enable provided callbacks for airflow DAGs and tasks by airflow_mcd package
    :param artifacts_export_enabled: whether to enable dbt artifacts export to Monte Carlo Data
    :param success_required: whether to require dbt artifacts export to MCD to be successful
    :param metastore_name: name of metastore in MCD
    """

    callbacks_enabled: bool = attrs.field(default=False)
    artifacts_export_enabled: bool = attrs.field(default=False)
    success_required: bool = attrs.field(default=False)
    metastore_name: str = attrs.field(default='')


@attrs.define(frozen=True)
class TableauIntegrationConfig:
    """
    Config for Tableau integration.
    There are two ways to authenticate:
        - with username and password
        - with token name and personal access token (PAT)
    You can use only one of them. If both are specified, then username and password will be used.

    :param server_address: The address of the Tableau Server or Tableau Online instance.

    :param username: The name of the user whose credentials will be used to sign in
    :param password: The password of the user.
    :param user_id_to_impersonate: Specifies the id (not the name) of the user to sign in as.

    :param token_name: The personal access token name.
    :param pat: The personal access token.

    :param site_id: This corresponds to the contentUrl attribute in the Tableau REST API.
        The site_id is the portion of the URL that follows the /site/ in the URL. For example,
        “MarketingTeam” is the site_id in the following URL MyServer/#/site/MarketingTeam/projects.
        To specify the default site on Tableau Server, you can use an empty string '' (single quotes, no space).
        For Tableau Cloud, you must provide a value for the site_id.
    """

    server_address: str = attrs.field(validator=attrs.validators.instance_of(str))

    username: Optional[str] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )
    password: Optional[str] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )
    user_id_to_impersonate: Optional[str] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )

    token_name: Optional[str] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )
    pat: Optional[str] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )

    site_id: Optional[str] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )


@attrs.define(frozen=True)
class DbtProjectConfig:
    """
    Config for dbt project.

    :param dbt_project_name: name of dbt project; refer to field `name` in dbt_project.yml
    # TODO: read from dbt_project.yml
    :param dbt_models_path: path to dbt models; it's used in k8s operators to find raw models code
    :param dbt_project_path: path to directory with dbt_project.yml
    :param dbt_profiles_path: path to directory with profiles.yml
    :param dbt_target_path: path to directory where compiled files (e.g. compiled models and tests) will be written
    :param dbt_log_path: path to directory where dbt logs will be written
    :param dbt_schema: dbt custom schema for models
    :additional_dbt_env: additional dbt env vars that will be passed to all dbt commands; they could be accessed in
        dbt project with `env_var` function; all keys and values must be strings

    """

    dbt_project_name: str = attrs.field(validator=attrs.validators.instance_of(str))
    dbt_models_path: str | Path = attrs.field(validator=attrs.validators.instance_of((str, Path)), converter=Path)
    dbt_project_path: str | Path = attrs.field(validator=attrs.validators.instance_of((str, Path)), converter=Path)
    dbt_profiles_path: str | Path = attrs.field(validator=attrs.validators.instance_of((str, Path)), converter=Path)
    dbt_target_path: str | Path = attrs.field(validator=attrs.validators.instance_of((str, Path)), converter=Path)
    dbt_log_path: str | Path = attrs.field(validator=attrs.validators.instance_of((str, Path)), converter=Path)
    dbt_schema: str = attrs.field(validator=attrs.validators.instance_of(str))
    additional_dbt_env: dict[str, str] = attrs.field(factory=dict, hash=True, eq=str)

    @additional_dbt_env.validator
    def validate_additional_dbt_env_type(self, attribute, value):
        for k, v in value.items():
            if not isinstance(k, str) or not isinstance(v, str):
                raise ValueError(f'additional_dbt_env must be a dict[str, str], got key={repr(k)}, value={repr(v)}')


@attrs.define(frozen=True)
class DbtDefaultTargetsConfig:
    """
    Config for default dbt targets for different operators.
    Target can be specified with:
        - explicit `dbt_af.dbt_target` field in model's config
        - fields `py_cluster`, `sql_cluster`, `daily_sql_cluster`, `bf_cluster` in model's config
        - default targets specified in this config

    :param default_target: default dbt target for all operators if other not specified explicitly; to set specific
        target for model please use `dbt_af.dbt_target` field in model's config
    :param default_for_tests_target: default dbt target for all dbt tests
    :param default_maintenance_target: default dbt target for all dbt maintenance tasks
    :param default_backfill_target: default dbt target for all dbt backfill tasks
    """

    default_target: str = attrs.field()
    default_for_tests_target: str = attrs.field(default=None)
    default_maintenance_target: str = attrs.field(default=None)
    default_backfill_target: str = attrs.field(default=None)

    def __attrs_post_init__(self):
        object.__setattr__(self, 'default_for_tests_target', self.default_for_tests_target or self.default_target)
        object.__setattr__(self, 'default_maintenance_target', self.default_maintenance_target or self.default_target)
        object.__setattr__(self, 'default_backfill_target', self.default_backfill_target or self.default_target)


@attrs.define(frozen=True)
class K8sConfig:
    """
    Config for k8s operators.

    :param airflow_identity_binding_selector: selector for airflow identity binding; if not specified, then
        no label aadpodidbinding will be used; it's azure specific label, please refer to docs for more details
        (https://learn.microsoft.com/en-us/azure/aks/use-azure-ad-pod-identity)
    """

    airflow_identity_binding_selector: str = attrs.field(default=None)


@attrs.define(frozen=True)
class RetryPolicy:
    """
    Config to configure airflow retries for one type of tasks
    """

    retries: int | None = None
    retry_delay: datetime.timedelta | None = None
    retry_exponential_backoff: bool | None = None
    max_retry_delay: datetime.timedelta | float | None = None

    def as_dict(self) -> dict[str, Any]:
        return {
            _attr.name: getattr(self, _attr.name)
            for _attr in attrs.fields(RetryPolicy)
            if getattr(self, _attr.name) is not None
        }


@attrs.define(frozen=True)
class RetriesConfig:
    """
    Config to parametrize different retry policies for all dbt-af dag components.
    Default policy must be configured or used with default values.
    Default policy will be used if task-specific policy is not specified.
    """

    default_retry_policy: RetryPolicy = attrs.field(
        default=RetryPolicy(
            retries=2,
            retry_delay=datetime.timedelta(minutes=1),
            retry_exponential_backoff=False,
        )
    )

    # dbt specific tasks
    dbt_run_retry_policy: RetryPolicy = attrs.field(default=None)
    dbt_seed_retry_policy: RetryPolicy = attrs.field(default=None)
    dbt_snapshot_retry_policy: RetryPolicy = attrs.field(default=None)
    dbt_test_retry_policy: RetryPolicy = attrs.field(default=None)
    dbt_source_freshness_retry_policy: RetryPolicy = attrs.field(default=None)

    # dbt-af specific tasks
    sensor_retry_policy: RetryPolicy = attrs.field(default=None)
    k8s_task_retry_policy: RetryPolicy = attrs.field(default=None)
    macros_retry_policy: RetryPolicy = attrs.field(default=None)
    supplemental_task_retry_policy: RetryPolicy = attrs.field(default=None)

    def __attrs_post_init__(self):
        """
        Substitute all unspecified policies with default one.
        If some values in task-specific policies are not specified, they will be used from default policy
        """
        for _attr in attrs.fields(RetriesConfig):
            _attr: attrs.Attribute
            if _attr.name == 'default_retry_policy':
                continue

            policy = getattr(self, _attr.name)
            policy: RetryPolicy | None
            # if a policy is completely unset, it's overwritten with default policy
            if policy is None:
                object.__setattr__(self, _attr.name, self.default_retry_policy)
                continue

            # if some parameters of the policy are not set, they are overwritten with values from default policy
            for _policy_attr in attrs.fields(RetryPolicy):
                _policy_attr: attrs.Attribute
                if getattr(policy, _policy_attr.name) is None:
                    object.__setattr__(policy, _policy_attr.name, getattr(self.default_retry_policy, _policy_attr.name))


@attrs.define(frozen=True)
class Config:
    """
    Main config for dbt-af.

    :param dbt_project: all dbt related params
    :param dbt_default_targets: default dbt targets for different operators
    :param dbt_executable_path: path to dbt executable to run tasks
    :param model_dependencies: section to parametrize how model dependencies are handled
    :param include_single_model_manual_dag: whether to include single model manual dag; it will create airflow dag
        without schedule, only with manual trigger and preset trigger form, where model name and date interval can be
        specified
    :param debug_mode_enabled: whether to run dbt commands with flag --debug
    :param retries_config: config with retries policies for different DAG component types
    :param max_active_dag_runs: max active dag runs for each airflow dag
    :param af_dag_description: description for airflow dags
    :param dag_start_date: default dag start date
    :param dry_run: A flag to enable or disable the execution of dbt commands and integrations with external tools.
        When set to `True`, dbt runs are skipped, and no changes will be applied to the database.
        This mode is typically used for testing or validating workflows without making changes to the target
        environment.
        Defaults to `False`, meaning dbt commands will execute as configured.
    :param use_dbt_target_specific_pools: whether to use dbt target specific pools; if True, then airflow pools will be
        created for each dbt target with pattern `dbt_{target_name}`; if False, then only the default pool will be used
    :param af_callbacks: config with callback functions for airflow DAGs and tasks
    :param mcd: config for mcd integration; must be installed as extra dependency
    :params tableau: config for Tableau integration
    :param k8s: settings for k8s operators

    :param is_dev: (deprecated) use `dry_run` instead
    """

    # dbt specific params
    dbt_project: DbtProjectConfig = attrs.field()

    # dbt-af specific params
    dbt_default_targets: DbtDefaultTargetsConfig = attrs.field()
    dbt_executable_path: str = attrs.field(default='dbt')
    model_dependencies: ModelDependenciesSection = attrs.field(factory=ModelDependenciesSection)
    include_single_model_manual_dag: bool = attrs.field(default=True)
    debug_mode_enabled: bool = attrs.field(default=True)

    # airflow-specific params
    retries_config: RetriesConfig = attrs.field(factory=RetriesConfig)
    max_active_dag_runs: int = attrs.field(default=50)
    af_dag_description: str = attrs.field(default='')
    dag_start_date: pendulum.datetime = attrs.field(default=pendulum.datetime(2023, 10, 1, 0, 0, 0, tz='UTC'))
    dry_run: bool = attrs.field(default=False)
    use_dbt_target_specific_pools: bool = attrs.field(default=True)

    # airflow callbacks config
    af_callbacks: Optional[CustomAfCallbacksConfig] = attrs.field(default=None)

    # Monte Carlo Data integration
    mcd: Optional[MCDIntegrationConfig] = attrs.field(default=None)

    # Tableau integration
    tableau: Optional[TableauIntegrationConfig] = attrs.field(default=None)

    # k8s
    k8s: K8sConfig = attrs.field(factory=K8sConfig)

    # DEPRECATED fields
    is_dev: bool = attrs.field(default=False)

    def __attrs_post_init__(self):
        # backward compatibility is_dev --> dry_run
        if self.is_dev and not self.dry_run:
            object.__setattr__(self, 'dry_run', self.is_dev)



import json
from typing import Optional

from airflow.models import DAG
from airflow.utils.task_group import TaskGroup

from dbt_af.conf import Config
from dbt_af.operators.base import DbtIntervalActionOperator
from dbt_af.parser.dbt_node_model import DbtAFMaintenanceConfig, DbtModelMaintenanceType


class DbtRunMacroOperation(DbtIntervalActionOperator):
    def __init__(self, dbt_af_config: Config, **kwargs):
        self.macro = self.macro_name

        super().__init__(
            task_id=self.af_task_name,
            dbt_af_config=dbt_af_config,
            retry_policy=dbt_af_config.retries_config.macros_retry_policy,
            **kwargs,
        )
        if self.macro_args:
            self.bash_options['--args'] = f"'{json.dumps(self.macro_args)}'"

    @property
    def af_task_name(self) -> str:
        safe_macro_name = f'dbt_{self.macro_name}'
        if hasattr(self, 'model_name'):
            return f'{safe_macro_name}__{self.model_name.replace(".", "_")}'
        return safe_macro_name

    @property
    def cli_command(self) -> str:
        return 'run-operation {macro}'

    @property
    def macro_name(self) -> str:
        raise NotImplementedError()

    @property
    def macro_args(self) -> Optional[dict]:
        return None


class DbtStageExternalSources(DbtRunMacroOperation):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @property
    def macro_name(self) -> str:
        return 'stage_external_sources'


class DbtPersistFullDocumentation(DbtRunMacroOperation):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @property
    def macro_name(self) -> str:
        return 'persist_docs_full'


class DbtPersistTableDocumentation(DbtRunMacroOperation):
    def __init__(self, model_name, maintenance_config: DbtAFMaintenanceConfig, **kwargs) -> None:
        self.model_name = model_name
        self.maintenance_config = maintenance_config

        super().__init__(**kwargs)

    @property
    def macro_name(self) -> str:
        return 'persist_docs'

    @property
    def macro_args(self) -> dict:
        return {'model_name': self.model_name}


class DbtOptimizeTable(DbtRunMacroOperation):
    def __init__(
        self,
        model_name: str,
        maintenance_config: DbtAFMaintenanceConfig,
        target_environment: str,
        max_active_tis_per_dag: int | None = 1,
        **kwargs,
    ) -> None:
        self.model_name = model_name
        self.maintenance_config = maintenance_config

        super().__init__(target_environment=target_environment, max_active_tis_per_dag=max_active_tis_per_dag, **kwargs)

    @property
    def macro_name(self) -> str:
        return 'optimize_table'

    @property
    def macro_args(self) -> dict:
        return {'model_name': self.model_name}


class DbtVacuumTable(DbtRunMacroOperation):
    def __init__(
        self,
        model_name: str,
        maintenance_config: DbtAFMaintenanceConfig,
        target_environment: str,
        max_active_tis_per_dag: int | None = 1,
        **kwargs,
    ) -> None:
        self.model_name = model_name
        self.maintenance_config = maintenance_config

        super().__init__(target_environment=target_environment, max_active_tis_per_dag=max_active_tis_per_dag, **kwargs)

    @property
    def macro_name(self) -> str:
        return 'vacuum_table'

    @property
    def macro_args(self) -> dict:
        return {'model_name': self.model_name}


class DbtDeduplicateTable(DbtRunMacroOperation):
    def __init__(
        self, model_name: str, maintenance_config: DbtAFMaintenanceConfig, target_environment: str, **kwargs
    ) -> None:
        self.retries = 0
        self.model_name = model_name
        self.maintenance_config = maintenance_config

        super().__init__(target_environment=target_environment, **kwargs)

    @property
    def macro_name(self) -> str:
        return 'deduplicate_table'

    @property
    def macro_args(self) -> dict:
        return {'model_name': self.model_name}


class DbtSetTTLOnTable(DbtRunMacroOperation):
    def __init__(
        self,
        model_name: str,
        maintenance_config: DbtAFMaintenanceConfig,
        target_environment: str,
        max_active_tis_per_dag: int = 1,
        **kwargs,
    ) -> None:
        self.model_name = model_name
        self.maintenance_config = maintenance_config

        super().__init__(target_environment=target_environment, max_active_tis_per_dag=max_active_tis_per_dag, **kwargs)

    @property
    def macro_name(self) -> str:
        return 'set_ttl_on_table'

    @property
    def macro_args(self) -> dict:
        return {
            'model_name': self.model_name,
            'key': self.maintenance_config.ttl.key,
            'expiration_timeout': self.maintenance_config.ttl.expiration_timeout,
            'additional_predicate': self.maintenance_config.ttl.additional_predicate,
            'force_predicate': self.maintenance_config.ttl.force_predicate,
        }


_MAPPING_MAINTENANCE_TYPE_TO_OPERATOR = {
    DbtModelMaintenanceType.PERSIST_DOCS: DbtPersistTableDocumentation,
    DbtModelMaintenanceType.OPTIMIZE_TABLES: DbtOptimizeTable,
    DbtModelMaintenanceType.VACUUM_TABLE: DbtVacuumTable,
    DbtModelMaintenanceType.DEDUPLICATE_TABLE: DbtDeduplicateTable,
    DbtModelMaintenanceType.SET_TTL_ON_TABLE: DbtSetTTLOnTable,
}


class DbtMaintenanceOperatorFactory:
    @staticmethod
    def create(
        model_name: str,
        maintenance_type: 'DbtModelMaintenanceType',
        task_group: 'TaskGroup',
        af_dag: 'DAG',
        maintenance_config: DbtAFMaintenanceConfig,
        dbt_af_config: 'Config',
        **kwargs,
    ) -> DbtRunMacroOperation:
        target_env = (
            dbt_af_config.dbt_default_targets.default_maintenance_target
            or dbt_af_config.dbt_default_targets.default_target
        )

        return _MAPPING_MAINTENANCE_TYPE_TO_OPERATOR[maintenance_type](
            model_name=model_name,
            maintenance_config=maintenance_config,
            task_group=task_group,
            dag=af_dag,
            max_active_tis_per_dag=1,
            target_environment=target_env,
            dbt_af_config=dbt_af_config,
            **kwargs,
        )



import logging
from typing import Any, Sequence

from airflow import __version__ as airflow_version
from airflow.operators.python import PythonVirtualenvOperator
from airflow.utils.context import Context
from packaging.version import Version

from dbt_af.conf import Config
from dbt_af.parser.dbt_profiles import VenvTarget


class DbtPythonVenvOperator(PythonVirtualenvOperator):
    template_fields: Sequence[str] = tuple(
        {'dbt_model_path', 'target_details'}.union(PythonVirtualenvOperator.template_fields)
    )
    template_fields_renderers = {'target_details': 'python'}

    def __init__(
        self,
        dbt_model_path: str,
        target_details: VenvTarget,
        dbt_af_config: Config,
        env: dict[str, str] | None = None,
        **kwargs,
    ):
        self.target_details = target_details
        self.env_vars = env
        self.dbt_model_path = dbt_model_path
        self.dbt_af_config = dbt_af_config

        def _python_callable(source_code: str) -> None:
            exec(source_code)

        if Version(airflow_version) >= Version('2.10.0'):
            kwargs |= {'inherit_env': target_details.inherit_env, 'env_vars': env}

        super().__init__(
            python_callable=_python_callable,
            requirements=target_details.requirements,
            system_site_packages=target_details.system_site_packages,
            python_version=target_details.python_version,
            pip_install_options=target_details.pip_install_options,
            index_urls=target_details.index_urls,
            **dbt_af_config.retries_config.k8s_task_retry_policy.as_dict(),
            **kwargs,
        )

    def execute(self, context: Context) -> Any:
        with open(self.dbt_af_config.dbt_project.dbt_models_path / self.dbt_model_path, 'r') as f:
            model_code = f.read()
        self.op_args = (model_code,)

        if Version(airflow_version) < Version('2.10.0') and self.env_vars:
            logging.error('Environment variables are not supported in Airflow versions below 2.10.0')

        return super().execute(context)



import logging
import os
import shutil
from functools import cached_property, partial
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Optional, Sequence

from airflow.hooks.subprocess import SubprocessHook
from airflow.models.dag import DAG
from airflow.sensors.external_task import ExternalTaskSensor
from airflow.sensors.python import PythonSensor
from airflow.utils.state import State

from dbt_af.common.af_scheduling_utils import (
    GLOBAL_TASK_SCHEDULE_MAPPINGS,
    _TaskScheduleMapping,
    calculate_task_to_wait_execution_date,
)
from dbt_af.common.constants import DBT_SENSOR_POOL
from dbt_af.common.scheduling import BaseScheduleTag, EScheduleTag
from dbt_af.conf import Config
from dbt_af.parser.dbt_node_model import WaitPolicy

if TYPE_CHECKING:
    from airflow.utils.task_group import TaskGroup

_DEFAULT_WAIT_TIMEOUT = 24 * 60 * 60
_DEFAULT_POKE_INTERVAL_SECONDS = 30 * 60
_RETRIES_COUNT = 30
_POKE_INTERVALS_SECONDS = {
    EScheduleTag.every15minutes.base_name: 30,
    EScheduleTag.hourly.base_name: 2 * 60,
    EScheduleTag.daily.base_name: 5 * 60,
    EScheduleTag.weekly.base_name: 30 * 60,
    EScheduleTag.monthly.base_name: 45 * 60,
}


def get_execution_date_fn_mapping(
    wait_policy: WaitPolicy,
    upstream_schedule_tag: BaseScheduleTag,
    downstream_schedule_tag: BaseScheduleTag,
) -> _TaskScheduleMapping:
    """
    This function returns mapping of functions to calculate the correct execution dates for sensors in runtime.
    It builds a whole matrix of all possible combinations for upstream and downstream schedules
    and calculates functions.
    """
    match wait_policy:
        case WaitPolicy.last:
            _mapping = GLOBAL_TASK_SCHEDULE_MAPPINGS[wait_policy]
            if _mapping.is_registered(upstream_schedule_tag, downstream_schedule_tag):
                return _mapping

            _mapping.add(
                upstream_schedule_tag,
                downstream_schedule_tag,
                partial(calculate_task_to_wait_execution_date),
            )
        case WaitPolicy.all:
            _mapping = GLOBAL_TASK_SCHEDULE_MAPPINGS[wait_policy]
            if _mapping.is_registered(upstream_schedule_tag, downstream_schedule_tag):
                return _mapping

            embeddings_number = downstream_schedule_tag.cron_expression().embeddings_number(
                upstream_schedule_tag.cron_expression(),
                is_upstream_bigger=downstream_schedule_tag < upstream_schedule_tag,
            )
            _mapping.add(
                upstream_schedule_tag,
                downstream_schedule_tag,
                (
                    [partial(calculate_task_to_wait_execution_date, num_iter=i) for i in range(embeddings_number)]
                    if embeddings_number
                    else [partial(calculate_task_to_wait_execution_date)]
                ),
            )
        case _:
            raise TypeError(f'Unknown wait policy {wait_policy}')

    return _mapping


class AfExecutionDateFn:
    """
    This class is used to get execution dates for sensors.
    Each function operates with execution date (aka logical date, start date)
    and waits for execution date of an upstream task.

    Example:
        daily -> hourly. Daily starts at 00:00 and waits for hourly at 23:00 of the previous day.
        It works like this: for daily task at 2023-01-02T00:00:00 execution date is 2023-01-01T00:00:00, and it will
        wait for hourly task with execution date at 2023-01-01T23:00:00 (it's data interval start!!!)
    """

    def __init__(
        self,
        upstream_schedule_tag: BaseScheduleTag,
        downstream_schedule_tag: BaseScheduleTag,
        wait_policy: WaitPolicy,
    ):
        self.upstream_schedule_tag = upstream_schedule_tag
        self.downstream_schedule_tag = downstream_schedule_tag
        self.wait_policy = wait_policy

    def get_execution_dates(self) -> list[Optional[callable]]:
        return get_execution_date_fn_mapping(
            self.wait_policy,
            self.upstream_schedule_tag,
            self.downstream_schedule_tag,
        ).get((self.upstream_schedule_tag, self.downstream_schedule_tag))


class DbtExternalSensor(ExternalTaskSensor):
    def __init__(
        self,
        dbt_af_config: Config,
        task_id: str,
        task_group: 'Optional[TaskGroup]',
        external_dag_id: str,
        external_task_id: str,
        execution_date_fn: callable,
        dep_schedule: BaseScheduleTag,
        dag: 'DAG',
        **kwargs,
    ) -> None:
        retry_policy = dbt_af_config.retries_config.sensor_retry_policy.as_dict()
        retry_policy['retries'] = max(_RETRIES_COUNT, retry_policy['retries'])
        super().__init__(
            task_id=task_id,
            task_group=task_group,
            external_dag_id=external_dag_id,
            external_task_id=external_task_id,
            execution_date_fn=execution_date_fn,
            dag=dag,
            max_active_tis_per_dag=None,
            pool=DBT_SENSOR_POOL if dbt_af_config.use_dbt_target_specific_pools else None,
            mode='reschedule',
            skipped_states=[State.NONE, State.SKIPPED],
            failed_states=[State.FAILED, State.UPSTREAM_FAILED],
            timeout=6 * 60 * 60,
            poke_interval=_POKE_INTERVALS_SECONDS.get(dep_schedule.base_name, _DEFAULT_POKE_INTERVAL_SECONDS),
            exponential_backoff=False,
            **retry_policy,
            **kwargs,
        )


class DbtSourceFreshnessSensor(PythonSensor):
    """
    :param wait_timeout: maximum time (in seconds) to wait for the sensor to return True
    :param retries: number of retries that should be performed before failing the sensor.
    :param poke_interval: time (in seconds) that the sensor should wait in between each try
    """

    template_fields: Sequence[str] = ('templates_dict', 'op_args', 'op_kwargs', 'env')
    template_fields_renderers = {'env': 'json'}

    def __init__(
        self,
        task_id: str,
        dag: 'DAG',
        env: dict,
        source_name: str,
        source_identifier: str,
        dbt_af_config: Config,
        target_environment: str = None,
        **kwargs,
    ):
        self.env = env
        self.source_name = source_name
        self.source_identifier = source_identifier
        self.target_environment = target_environment or dbt_af_config.dbt_default_targets.default_for_tests_target
        self.dbt_af_config = dbt_af_config

        retry_policy = dbt_af_config.retries_config.dbt_source_freshness_retry_policy.as_dict()
        if retries := retry_policy.get('retries'):
            logging.debug(
                'For %s number of retries are dynamically calculated and passed value retries=%s will be ignored',
                self.__class__.__name__,
                retries,
            )

        _poke_interval = (
            retry_policy.get('retry_delay').total_seconds()
            if retry_policy.get('retry_delay')
            else _DEFAULT_POKE_INTERVAL_SECONDS
        )
        retry_policy['retries'] = _DEFAULT_WAIT_TIMEOUT // _poke_interval - 1
        retry_policy['poke_interval'] = _poke_interval
        retry_policy['timeout'] = _DEFAULT_WAIT_TIMEOUT

        super().__init__(
            task_id=task_id,
            dag=dag,
            pool=DBT_SENSOR_POOL if dbt_af_config.use_dbt_target_specific_pools else None,
            mode='reschedule',
            python_callable=self._check_freshness,
            **retry_policy,
            **kwargs,
        )

    @cached_property
    def subprocess_hook(self):
        """Returns hook for running the bash command."""
        return SubprocessHook()

    def _check_freshness(self):
        env = os.environ.copy()
        env.update(self.env)
        with TemporaryDirectory(dir=self.dbt_af_config.dbt_project.dbt_target_path) as tmp_target_path:
            shutil.copy(
                self.dbt_af_config.dbt_project.dbt_project_path / 'target/manifest.json',
                f'{tmp_target_path}/manifest.json',
            )
            freshness_cmd = ' && '.join(
                [
                    f'{self.dbt_af_config.dbt_executable_path} source freshness '
                    f'{"--debug" if self.dbt_af_config.debug_mode_enabled else ""} '
                    f'{"-h" if self.dbt_af_config.dry_run else ""} '
                    f'--target-path {tmp_target_path} '
                    f'--profiles-dir $DBT_PROFILES_DIR '
                    f'--project-dir $DBT_PROJECT_DIR '
                    f'--target {self.target_environment} '
                    f'--select source:{self.source_name}.{self.source_identifier}',
                ]
            )
            result = self.subprocess_hook.run_command(
                command=['bash', '-c', freshness_cmd],
                env=env,
                cwd=str(self.dbt_af_config.dbt_project.dbt_project_path),
            )
        if result.exit_code:
            return False

        return True



from collections import defaultdict
from typing import Optional

import pendulum

from dbt_af.builder.backfill_dag_components import BackfillDagModel, BackfillDagSnapshot
from dbt_af.builder.dag_components import DagComponent, DagModel, DagSeed, DagSnapshot, LargeTest, MediumTests
from dbt_af.builder.domain_dag import BackfillDomainDag, DomainDag, DomainDagFactory, DomainDagType
from dbt_af.builder.maintenance_dag_components import MaintenanceDagComponent
from dbt_af.common.constants import DOMAIN_DAG_START_DATE_FMT
from dbt_af.conf import Config
from dbt_af.parser.dbt_node_model import DbtModelMaintenanceType, DbtNode
from dbt_af.parser.dbt_profiles import Profiles
from dbt_af.parser.dbt_source_model import DbtSource


class DagComponentFactory:
    @staticmethod
    def create(dbt_node: DbtNode, domain_dag: DomainDag, backfill: bool = False) -> DagComponent:
        if dbt_node.is_model():
            return DagModel(dbt_node, domain_dag) if not backfill else BackfillDagModel(dbt_node, domain_dag)
        if dbt_node.is_snapshot():
            return DagSnapshot(dbt_node, domain_dag) if not backfill else BackfillDagSnapshot(dbt_node, domain_dag)
        if dbt_node.is_large_test() and not backfill:
            return LargeTest(dbt_node.resource_name, domain_dag, dbt_node.config)
        if dbt_node.is_seed():
            return DagSeed(dbt_node, domain_dag)

        raise ValueError(f'Unknown node type: {dbt_node} for {"backfill" if backfill else ""} {domain_dag}')


class DomainDagsRegistry:
    def __init__(self, config: Config, dags_type: DomainDagType = DomainDagType.SCHEDULED):
        self.config = config
        self.dags_type = dags_type

        self._domain_dags: dict[str, DomainDag] = {}

    def clear(self):
        self._domain_dags = {}

    def _get_domain_dag_hash(self, dbt_node: DbtNode) -> str:
        if dbt_node.is_large_test():
            return f'{dbt_node.domain}__large_tests'
        if self.dags_type == DomainDagType.BACKFILL:
            return f'{dbt_node.domain}__bf'
        if self.dags_type == DomainDagType.SCHEDULED:
            return f'{dbt_node.domain}_{dbt_node.config.schedule.safe_name}'
        if self.dags_type == DomainDagType.MAINTENANCE:
            return f'{dbt_node.domain}__maintenance'

        raise TypeError(f'Unknown dag type: {self.dags_type}')

    def get(self, dbt_node: DbtNode) -> DomainDag:
        domain_dag_name = self._get_domain_dag_hash(dbt_node)

        if domain_dag_name not in self._domain_dags:
            self._domain_dags[domain_dag_name] = DomainDagFactory.create(
                dag_type=self.dags_type if not dbt_node.is_large_test() else DomainDagType.LARGE_TESTS,
                domain_name=dbt_node.domain,
                schedule=dbt_node.config.schedule,
                config=self.config,
            )

        return self._domain_dags[domain_dag_name]


class DbtAfGraph:
    def __init__(self, nodes: list[DbtNode], sources: list[DbtSource], config: Config):
        self.config = config

        self.dbt_nodes: list[DbtNode] = nodes
        self.dbt_sources: list[DbtSource] = sources

        # dbt-af components
        self._domain_dags_registry = DomainDagsRegistry(config=self.config)
        self._domain_bf_dags_registry = DomainDagsRegistry(config=self.config, dags_type=DomainDagType.BACKFILL)
        self._domain_maintenance_dags_registry = DomainDagsRegistry(
            config=self.config,
            dags_type=DomainDagType.MAINTENANCE,
        )
        self._models: dict[str, DagModel] = {}
        self._large_tests: dict[str, LargeTest] = {}
        self._dag_components_registry: dict[str, DagComponent] = {}
        self._medium_tests: dict[DomainDag, MediumTests] = {}
        self._maintenance_components: dict[DomainDag, dict[DbtModelMaintenanceType, MaintenanceDagComponent]] = (
            defaultdict(dict)
        )

        self.nodes: list[DagComponent] = []

    @classmethod
    def from_manifest(
        cls,
        manifest: dict,
        profiles: dict,
        project_profile_name: str,
        config: Config,
        etl_service_name: Optional[str] = None,
    ) -> 'DbtAfGraph':
        project_profile = Profiles(**profiles)[project_profile_name]

        nodes = []
        for node_info in manifest['nodes'].values():
            node = DbtNode(**node_info)
            node.set_target_details(project_profile, config.dbt_default_targets)
            if node.resource_type in ('test', 'model', 'snapshot', 'seed'):
                # TODO: add sensors for models in different etl services
                if etl_service_name and not node.is_at_etl_service(etl_service_name):
                    continue
                nodes.append(node)

        sources = [DbtSource(**source_info) for source_info in manifest['sources'].values()]

        graph = cls(nodes, sources, config)
        graph._build_dags()
        return graph

    def clear_registries(self):
        self._domain_dags_registry.clear()
        self._domain_bf_dags_registry.clear()
        self._models = {}
        self._large_tests = {}
        self._dag_components_registry = {}
        self._medium_tests = {}

    def _build_dags(self):
        dag_components = self._build_dag_components(self.dbt_nodes)
        self.clear_registries()
        backfill_dag_components = self._build_backfill_dag_components(self.dbt_nodes)

        self.nodes = dag_components + backfill_dag_components

    def _find_parent_node_for_test(self, node: DbtNode) -> DagModel | BackfillDomainDag:
        original_file_path = node.original_file_path_without_extension
        for upstream in node.depends_on:
            if self._models[upstream].dbt_node.original_file_path_without_extension == original_file_path:
                return self._models[upstream]

        raise ValueError(f'Could not find parent node for medium test {node.unique_id}')

    def _collect_all_models(self, nodes, backfill: bool = False) -> None:
        domain_dags_registry = self._domain_bf_dags_registry if backfill else self._domain_dags_registry
        for node in nodes:
            domain_dag = domain_dags_registry.get(node)
            try:
                self._dag_components_registry[node.unique_id] = DagComponentFactory.create(
                    dbt_node=node,
                    domain_dag=domain_dag,
                    backfill=backfill,
                )
            except ValueError:
                # node is not an independent dag component or should not be in dag
                continue

        self._models = {
            node.unique_id: self._dag_components_registry[node.unique_id]
            for node in nodes
            if node.is_model() or node.is_seed()
        }
        if not backfill:
            self._large_tests = {
                node.unique_id: self._dag_components_registry[node.unique_id] for node in nodes if node.is_large_test()
            }

    def _collect_maintenance_components(self) -> None:
        for model in self._models.values():
            if maintenance_types := model.dbt_node.get_required_maintenance_types():
                domain_dag = self._domain_maintenance_dags_registry.get(model.dbt_node)
                for maintenance_type in maintenance_types:
                    if maintenance_type not in self._maintenance_components[domain_dag]:
                        self._maintenance_components[domain_dag][maintenance_type] = MaintenanceDagComponent(
                            domain_dag=domain_dag,
                            maintenance_type=maintenance_type,
                            node_config=model.dbt_node.config,
                        )
                    self._maintenance_components[domain_dag][maintenance_type].add_model(model.dbt_node)

    def _resolve_dependencies(self, nodes: list[DbtNode], backfill: bool = False) -> None:
        sources: dict[str, DbtSource] = {source.unique_id: source for source in self.dbt_sources}

        # set dependencies for models and snapshots
        for node in nodes:
            if node.is_model() or node.is_snapshot():
                for upstream in node.depends_on:
                    self._models[node.unique_id].add_dependency(self._models[upstream])
                for upstream in node.depends_on_sources:
                    self._models[node.unique_id].add_source_dependency(sources[upstream])

            elif node.is_small_test():
                for upstream in node.depends_on:
                    self._models[upstream].add_small_test(node.resource_name)

            elif node.is_medium_test():
                parent_node = self._find_parent_node_for_test(node)
                parent_domain_dag = parent_node.domain_dag
                if parent_domain_dag not in self._medium_tests:
                    self._medium_tests[parent_domain_dag] = MediumTests(parent_domain_dag, parent_node.dbt_node.config)
                self._medium_tests[parent_domain_dag].add_test(MediumTests.get_medium_test_name(node, parent_node))

            elif node.is_large_test() and not backfill:
                # set dependencies for large tests only for regular scheduled dags
                for upstream in node.depends_on:
                    self._large_tests[node.unique_id].add_dependency(self._models[upstream])

    def _bind_medium_tests(self) -> None:
        """
        Bind medium tests to all model nodes in domain so medium tests will be executed after all models
        """
        for model in self._models.values():
            if model.domain_dag in self._medium_tests:
                self._medium_tests[model.domain_dag].add_dependency(model)

    def _build_dag_components(self, nodes: list[DbtNode]) -> list[DagComponent]:
        self._collect_all_models(nodes)
        self._collect_maintenance_components()

        self._resolve_dependencies(nodes)

        self._bind_medium_tests()

        return (
            list(self._models.values())
            + list(self._medium_tests.values())
            + list(self._large_tests.values())
            + list(*[maintenance.values() for maintenance in self._maintenance_components.values()])
        )

    def _build_backfill_dag_components(self, nodes: list[DbtNode]) -> list[DagComponent]:
        self._collect_all_models(nodes, backfill=True)

        self._resolve_dependencies(nodes, backfill=True)

        self._bind_medium_tests()

        return list(self._models.values()) + list(self._medium_tests.values())


def get_domain_dag_start_date(graph: DbtAfGraph, domain_dag: DomainDag) -> pendulum.datetime:
    min_date_from_nodes = min(
        node.node_config.domain_start_date for node in graph.nodes if node.domain_dag == domain_dag
    )
    if min_date_from_nodes:
        return pendulum.from_format(min_date_from_nodes, DOMAIN_DAG_START_DATE_FMT)
    return graph.config.dag_start_date
